@@ -1,4 +1,5 @@
-from sanic import Sanic
+from sanic import Sanic  # type: ignore
+
 from sanic_testing.testing import SanicASGITestClient, SanicTestClient
 
 
