@@ -1,15 +1,17 @@
 import typing
+from functools import partial
 from json import JSONDecodeError
 from socket import socket
+from types import SimpleNamespace
 
 import httpx
 import websockets
-from sanic import Sanic
-from sanic.asgi import ASGIApp
-from sanic.exceptions import MethodNotSupported
-from sanic.log import logger
-from sanic.request import Request
-from sanic.response import HTTPResponse, text
+from sanic import Sanic  # type: ignore
+from sanic.asgi import ASGIApp  # type: ignore
+from sanic.exceptions import MethodNotSupported  # type: ignore
+from sanic.log import logger  # type: ignore
+from sanic.request import Request  # type: ignore
+from sanic.response import HTTPResponse, text  # type: ignore
 
 ASGI_HOST = "mockserver"
 ASGI_PORT = 1234
@@ -28,14 +30,16 @@ class SanicTestClient:
         self.app = app
         self.port = port
         self.host = host
+        app.listener("after_server_start")(self._start_test_mode)
+        app.listener("before_server_stop")(self._end_test_mode)
 
-        @app.listener("after_server_start")
-        def _start_test_mode(sanic, *args, **kwargs):
-            sanic.test_mode = True
+    @classmethod
+    def _start_test_mode(cls, sanic, *args, **kwargs):
+        sanic.test_mode = True
 
-        @app.listener("before_server_end")
-        def _end_test_mode(sanic, *args, **kwargs):
-            sanic.test_mode = False
+    @classmethod
+    def _end_test_mode(cls, sanic, *args, **kwargs):
+        sanic.test_mode = False
 
     def get_new_session(self, **kwargs) -> httpx.AsyncClient:
         return httpx.AsyncClient(verify=False, **kwargs)
@@ -46,22 +50,24 @@ class SanicTestClient:
         session_kwargs = kwargs.pop("session_kwargs", {})
 
         if method == "websocket":
+            ws_proxy = SimpleNamespace()
             async with websockets.connect(url, *args, **kwargs) as websocket:
-                websocket.opened = websocket.open
-                return websocket
+                ws_proxy.ws = websocket
+                ws_proxy.opened = True
+            return ws_proxy
         else:
             async with self.get_new_session(**session_kwargs) as session:
 
                 try:
                     if method == "request":
-                        args = [url] + list(args)
+                        args = tuple([url] + list(args))
                         url = kwargs.pop("http_method", "GET").upper()
                     response = await getattr(session, method.lower())(
                         url, *args, **kwargs
                     )
                 except httpx.HTTPError as e:
                     if hasattr(e, "response"):
-                        response = e.response
+                        response = getattr(e, "response")
                     else:
                         logger.error(
                             f"{method.upper()} {url} received no response!",
@@ -88,6 +94,38 @@ class SanicTestClient:
 
             return response
 
+    @classmethod
+    def _collect_request(cls, results, request):
+        if results[0] is None:
+            results[0] = request
+
+    async def _collect_response(
+        self,
+        method,
+        url,
+        exceptions,
+        results,
+        sanic,
+        loop,
+        **request_kwargs,
+    ):
+        try:
+            response = await self._local_request(method, url, **request_kwargs)
+            results[-1] = response
+            if method == "websocket":
+                await response.ws.close()
+        except Exception as e:
+            logger.exception("Exception")
+            exceptions.append(e)
+        finally:
+            self.app.stop()
+
+    async def _error_handler(self, request, exception):
+        if request.method in ["HEAD", "PATCH", "PUT", "DELETE"]:
+            return text("", exception.status_code, headers=exception.headers)
+        else:
+            return self.app.error_handler.default(request, exception)
+
     def _sanic_endpoint_test(
         self,
         method: str = "get",
@@ -98,28 +136,17 @@ class SanicTestClient:
         host: str = None,
         *request_args,
         **request_kwargs,
-    ) -> typing.Tuple[typing.Union[Request, HTTPResponse]]:
+    ) -> typing.Union[typing.Tuple[Request, HTTPResponse], HTTPResponse]:
         results = [None, None]
-        exceptions = []
+        exceptions: typing.List[Exception] = []
 
         server_kwargs = server_kwargs or {"auto_reload": False}
+        _collect_request = partial(self._collect_request, results)
 
         if gather_request:
-
-            def _collect_request(request):
-                if results[0] is None:
-                    results[0] = request
-
             self.app.request_middleware.appendleft(_collect_request)
 
-        @self.app.exception(MethodNotSupported)
-        async def error_handler(request, exception):
-            if request.method in ["HEAD", "PATCH", "PUT", "DELETE"]:
-                return text(
-                    "", exception.status_code, headers=exception.headers
-                )
-            else:
-                return self.app.error_handler.default(request, exception)
+        self.app.exception(MethodNotSupported)(self._error_handler)
 
         if self.port:
             server_kwargs = dict(
@@ -147,18 +174,16 @@ class SanicTestClient:
         # known until this function is called, so fix that here
         url = url.replace(":None/", f":{port}/")
 
-        @self.app.listener("after_server_start")
-        async def _collect_response(sanic, loop):
-            try:
-                response = await self._local_request(
-                    method, url, *request_args, **request_kwargs
-                )
-                results[-1] = response
-            except Exception as e:
-                logger.exception("Exception")
-                exceptions.append(e)
-            finally:
-                self.app.stop()
+        self.app.listener("after_server_start")(
+            partial(
+                self._collect_response,
+                method,
+                url,
+                exceptions,
+                results,
+                **request_kwargs,
+            )
+        )
 
         self.app.run(debug=debug, **server_kwargs)
         self.app.listeners["after_server_start"].pop()
@@ -167,6 +192,11 @@ class SanicTestClient:
             raise ValueError(f"Exception during request: {exceptions}")
 
         if gather_request:
+            try:
+                self.app.request_middleware.remove(_collect_request)
+            except BaseException:  # noqa
+                pass
+
             try:
                 request, response = results
                 return request, response
@@ -226,6 +256,7 @@ class SanicASGITestClient(httpx.AsyncClient):
         base_url: str = ASGI_BASE_URL,
         suppress_exceptions: bool = False,
     ) -> None:
+
         app.__class__.__call__ = app_call_with_return
         app.asgi = True
 
@@ -235,20 +266,25 @@ class SanicASGITestClient(httpx.AsyncClient):
 
         super().__init__(transport=transport, base_url=base_url)
 
+        self.gather_request = True
         self.last_request = None
 
-        def _collect_request(request):
+        app.listener("after_server_start")(self._start_test_mode)
+        app.listener("before_server_stop")(self._end_test_mode)
+
+    def _collect_request(self, request):
+        if self.gather_request:
             self.last_request = request
+        else:
+            self.last_request = None
 
-        @app.listener("after_server_start")
-        def _start_test_mode(sanic, *args, **kwargs):
-            sanic.test_mode = True
+    @classmethod
+    def _start_test_mode(cls, sanic, *args, **kwargs):
+        sanic.test_mode = True
 
-        @app.listener("before_server_end")
-        def _end_test_mode(sanic, *args, **kwargs):
-            sanic.test_mode = False
-
-        app.request_middleware.appendleft(_collect_request)
+    @classmethod
+    def _end_test_mode(cls, sanic, *args, **kwargs):
+        sanic.test_mode = False
 
     async def request(self, method, url, gather_request=True, *args, **kwargs):
 
@@ -259,13 +295,25 @@ class SanicASGITestClient(httpx.AsyncClient):
             scheme = "ws" if method == "websocket" else "http"
             url = f"{scheme}://{ASGI_HOST}:{ASGI_PORT}{url}"
 
+        if self._collect_request not in self.sanic_app.request_middleware:
+            self.sanic_app.request_middleware.appendleft(self._collect_request)
+
         self.gather_request = gather_request
         response = await super().request(method, url, *args, **kwargs)
         response.status = response.status_code
         response.body = response.content
         response.content_type = response.headers.get("content-type")
+        if gather_request:
+            return self.last_request, response
+        return response
 
-        return self.last_request, response
+    @classmethod
+    async def _ws_receive(cls):
+        return {}
+
+    @classmethod
+    async def _ws_send(cls, message):
+        pass
 
     async def websocket(self, uri, subprotocols=None, *args, **kwargs):
         scheme = "ws"
@@ -290,14 +338,27 @@ class SanicASGITestClient(httpx.AsyncClient):
             "root_path": root_path,
             "path": path,
             "query_string": b"",
+            "subprotocols": subprotocols,
         }
 
-        async def receive():
-            return {}
+        await self.sanic_app(scope, self._ws_receive, self._ws_send)
 
-        async def send(message):
+        return None, {"opened": True}
+
+    def __getstate__(self):
+        # Cookies cannot be pickled, because they contain a ThreadLock
+        try:
+            del self._cookies
+        except AttributeError:
             pass
+        return self.__dict__
 
-        await self.sanic_app(scope, receive, send)
-
-        return None, {}
+    def __setstate__(self, d):
+        try:
+            del d["_cookies"]
+        except LookupError:
+            pass
+        self.__dict__.update(d)
+        # Need to create a new CookieJar when unpickling,
+        # because it was killed on Pickle
+        self._cookies = httpx.Cookies()
